// Copyright 2011 Google Inc. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package com.google.api.ads.dfp.lib.client;

import com.google.api.ads.common.lib.auth.ClientLoginCompatible;
import com.google.api.ads.common.lib.auth.OAuth2Compatible;
import com.google.api.ads.common.lib.auth.OAuthCompatible;
import com.google.api.ads.common.lib.client.AdsSession;
import com.google.api.ads.common.lib.client.Endpoint;
import com.google.api.ads.common.lib.conf.ConfigurationHelper;
import com.google.api.ads.common.lib.conf.ConfigurationLoadException;
import com.google.api.ads.common.lib.exception.ValidationException;
import com.google.api.client.auth.oauth.OAuthParameters;
import com.google.api.client.auth.oauth2.Credential;
import com.google.common.annotations.VisibleForTesting;
import com.google.common.base.Preconditions;

import org.apache.commons.configuration.Configuration;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;


/**
 * A {@code DfpSession} represents a single session of DFP use.
 *
 * @author Adam Rogal
 */
public class DfpSession implements AdsSession, OAuthCompatible, OAuth2Compatible,
    ClientLoginCompatible {

  /**
   * Enum representing the endpoint server, aka environment.
   *
   * @deprecated
   */
  @Deprecated
  public static enum Environment implements Endpoint {
    PRODUCTION("https://www.google.com/");

    private final String endpoint;

    private Environment(String endpoint) {
      this.endpoint = endpoint;
    }

    /**
     * Gets the endpoint.
     */
    public String getEndpoint() {
      return endpoint;
    }
  }

  private String endpoint;
  private String clientLoginToken;
  private String applicationName;
  private String networkCode;
  private OAuthParameters oAuthParameters;
  private Credential oAuth2Credential;


  public static final String DEFAULT_ENDPOINT = "https://www.google.com/";

  /**
   * Private constructor.
   */
  private DfpSession() {}

  /**
   * Copy constructor.
   */
  private DfpSession(DfpSession copy) {
    this.applicationName = copy.applicationName;
    this.clientLoginToken = copy.clientLoginToken;
    this.endpoint = copy.endpoint;
    this.networkCode = copy.networkCode;
    this.oAuth2Credential = copy.oAuth2Credential;
    this.oAuthParameters = copy.oAuthParameters;
  }

  /**
   * Gets the application name.
   */
  public String getApplicationName() {
    return applicationName;
  }

  /**
   * Gets the OAuth parameters.
   */
  public OAuthParameters getOAuthParameters() {
    return oAuthParameters;
  }

  /**
   * Sets the OAuth parameters. Any other authentication credentials on the
   * session will be removed.
   */
  public void setOAuthParameters(OAuthParameters oAuthParameters) {
    Preconditions.checkNotNull(oAuthParameters, "oAuthParameters cannot be null.");
    clearAuthentication();
    this.oAuthParameters = oAuthParameters;
  }


  /**
   * Gets the OAuth2 credentials.
   */
  public Credential getOAuth2Credential() {
    return oAuth2Credential;
  }

  /**
   * Sets the OAuth2 credential. Any other authentication credentials on the
   * session will be removed.
   */
  public void setOAuth2Credential(Credential oAuth2Credential) {
    Preconditions.checkNotNull(oAuth2Credential, "oAuth2Credential cannot be null.");
    clearAuthentication();
    this.oAuth2Credential = oAuth2Credential;
  }

  /**
   * Gets the client login token.
   */
  public String getClientLoginToken() {
    return clientLoginToken;
  }

  /**
   * Sets the client login token. Any other authentication credentials on the
   * session will be removed.
   */
  public void setClientLoginToken(String clientLoginToken) {
    Preconditions.checkNotNull(clientLoginToken, "clientLoginToken cannot be null.");
    clearAuthentication();
    this.clientLoginToken = clientLoginToken;
  }

  /**
   * Gets the endpoint.
   */
  public String getEndpoint() {
    return endpoint;
  }

  /**
   * Determines if the session is for the supplied
   * {@link DfpSession.Environment}.
   *
   * @deprecated
   */
  @Deprecated
  public boolean isEnvironment(Endpoint environment) {
    return environment.getEndpoint().equals(this.endpoint);
  }

  /**
   * Gets the network code.
   */
  public String getNetworkCode() {
    return networkCode;
  }

  /**
   * Sets the network code.
   */
  public void setNetworkCode(String networkCode) {
    this.networkCode = networkCode;
  }

  /**
   * Clears all the authentication credentials from this session.
   */
  private void clearAuthentication() {
    oAuth2Credential = null;
    oAuthParameters = null;
    clientLoginToken = null;
  }


  /**
   * Builder for {@code DfpSession}.
   */
  public static class Builder implements com.google.api.ads.common.lib.utils.Builder<DfpSession> {

    private DfpSession dfpSession;
    private String environmentString;
    @VisibleForTesting
    ConfigurationHelper configHelper;

    /**
     * Constructor.
     */
    public Builder() {
      this.dfpSession = new DfpSession();
      this.configHelper = new ConfigurationHelper();
    }

    @VisibleForTesting
    Builder(ConfigurationHelper configHelper) {
      this.dfpSession = new DfpSession();
      this.configHelper = configHelper;
    }

    public Builder fromFile() throws ConfigurationLoadException {
      return fromFile(Builder.DEFAULT_CONFIGURATION_FILENAME);
    }

    public Builder fromFile(String path) throws ConfigurationLoadException {
      return from(configHelper.fromFile(path));
    }

    public Builder fromFile(File path) throws ConfigurationLoadException {
      return from(configHelper.fromFile(path));
    }

    public Builder fromFile(URL path) throws ConfigurationLoadException {
      return from(configHelper.fromFile(path));
    }

    /**
     * Reads properties from the provided {@link Configuration} object.<br>
     * <br>
     * Known properties:
     * <ul>
     * <li>api.dfp.clientLoginToken</li>
     * <li>api.dfp.applicationName</li>
     * <li>api.dfp.networkCode</li>
     * <li>api.dfp.endpoint</li>
     * <li>Deprecated: api.dfp.environment - e.g. production</li>
     * </ul>
     *
     * @param config the configuration
     * @return Builder populated from the configuration
     */
    public Builder from(Configuration config) {
      dfpSession.clientLoginToken = config.getString("api.dfp.clientLoginToken", null);
      dfpSession.applicationName = config.getString("api.dfp.applicationName", null);
      dfpSession.networkCode = config.getString("api.dfp.networkCode", null);
      dfpSession.endpoint = config.getString("api.dfp.endpoint", null);
      environmentString = config.getString("api.dfp.environment", null);

      return this;
    }

    /**
     * Includes hard-coded ClientLogin token that will be used instead of
     * fetching a new one.
     */
    public Builder withClientLoginToken(String clientLoginToken) {
      dfpSession.clientLoginToken = clientLoginToken;
      return this;
    }

    /**
     * Includes OAuth parameters to be used for OAuth authentication.
     */
    public Builder withOAuthParameters(OAuthParameters oAuthParameters) {
      dfpSession.oAuthParameters = oAuthParameters;
      return this;
    }

    /**
     * Includes OAuth2 credential to be used for OAuth2 authentication.
     */
    public Builder withOAuth2Credential(Credential oAuth2Credential) {
      dfpSession.oAuth2Credential = oAuth2Credential;
      return this;
    }

    /**
     * Includes network code. Required for most service calls.
     */
    public Builder withNetworkCode(String networkCode) {
      dfpSession.networkCode = networkCode;
      return this;
    }

    /**
     * Includes application name (any string of your choice). Required.
     */
    public Builder withApplicationName(String applicationName) {
      dfpSession.applicationName = applicationName;
      return this;
    }

    /**
     * Includes the specified environment, e.g. {@link Environment#PRODUCTION}.
     * Optional and defaults to {@link Environment#PRODUCTION}.
     *
     * The environment and endpoint in DFP are currently synonymous. Setting an
     * environment will clear out any endpoint you have previously set.
     *
     * @deprecated
     */
    @Deprecated
    public Builder withEnvironment(Environment environment) {
      dfpSession.endpoint = environment.getEndpoint();
      environmentString = null;
      return this;
    }

    /**
     * Override the endpoint server. Optional and defaults to
     * https://www.google.com/.
     *
     * The environment and endpoint in DFP are currently synonymous.
     * Setting an endpoint will clear out any environment you have previously set.
     */
    public Builder withEndpoint(String endpoint) {
      dfpSession.endpoint = endpoint;
      environmentString = null;
      return this;
    }


    /**
     * Builds the {@code DfpSession}.
     *
     * @return the built {@code DfpSession}
     * @throws ValidationException if the {@code DfpSession} did not validate
     */
    public DfpSession build() throws ValidationException {
      defaultOptionals();
      validate(dfpSession);
      return new DfpSession(dfpSession);
    }

    /**
     * Fills in defaults if {@code null}.
     */
    private void defaultOptionals() {
      if (dfpSession.endpoint == null) {
        dfpSession.endpoint = DEFAULT_ENDPOINT;
      }
    }

    /**
     * Validates the DFP session.
     */
    private void validate(DfpSession dfpSession) throws ValidationException {
      if (environmentString != null) {
        try {
          withEnvironment(Environment.valueOf(environmentString.toUpperCase()));
        } catch (IllegalArgumentException e) {
          throw new ValidationException(String.format("Environment [%s] not recognized.",
              environmentString), "api.dfp.environment", e);
        }
      }

      // Check for valid OAuth or ClientLogin information.
      if (dfpSession.getClientLoginToken() != null
          && (dfpSession.getOAuthParameters() != null
              || dfpSession.getOAuth2Credential() != null)) {
        throw new ValidationException("Both ClientLogin authetication (clientLoginToken) and"
            + " OAuth or OAuth2 authentication cannot be used at the same time.", "");
      }

      // Check for valid OAuth/2 information.
      if (dfpSession.getClientLoginToken() == null
          && (dfpSession.getOAuthParameters() != null
              && dfpSession.getOAuth2Credential() != null)) {
        throw new ValidationException(
            "Both OAuth and OAuth2 cannot be used at the same time.", "");
      }

      // Check for at least one authentication mechanism.
      if (dfpSession.getClientLoginToken() == null && dfpSession.getOAuthParameters() == null
          && dfpSession.getOAuth2Credential() == null) {
        throw new ValidationException(
            "Either ClientLogin authetication, OAuth, or OAuth2 authentication must be used.", "");
      }


      // Make sure they specify a valid environment.
      try {
        new URL(dfpSession.getEndpoint());
      } catch (MalformedURLException e) {
        throw new ValidationException(String.format("Endpoint [%s] not recognized as a valid URL.",
            dfpSession.getEndpoint()), "endpoint", e);
      }
    }
  }
}
